VAGRANT_TEMPLATE = '''# -*- mode: ruby -*-
# vi: set ft=ruby :

VAGRANTFILE_API_VERSION = "2"

Vagrant.configure(VAGRANTFILE_API_VERSION) do |config|

  config.vm.box = "{box}"

end'''

DOCKER_TEMPLATE = '''FROM {image}

# install packages
RUN apt-get update && \\
    NEEDRESTART_MODE=a apt-get install -y {packages}

# init user and ssh
EXPOSE 22
RUN useradd --create-home --shell /bin/bash -g sudo {user}
RUN chown -R vagd:sudo /home/vagd
RUN chmod u+s /usr/bin/sudo
RUN echo "vagd ALL=(ALL) NOPASSWD:ALL" > /etc/sudoers.d/vagd && chmod 0440 /etc/sudoers.d/vagd
USER {user}

WORKDIR /home/{user}
COPY {keyfile} .ssh/authorized_keys

USER root
RUN mkdir -p /run/sshd && \\
    chmod 755 /run/sshd
    
    
CMD /usr/sbin/sshd; \\
    while true; do sleep 1m; done
'''

# TODO: proper template generation for alpine
DOCKER_ALPINE_TEMPLATE = '''FROM {image}

# install packages
RUN apk update
RUN apk add --no-cache python3
RUN apk add --no-cache musl-dbg
# install gdb
RUN apk add --no-cache gdb
# install ssh server support and keys
RUN apk add --no-cache openssh
# install sudo
RUN apk add --no-cache sudo

EXPOSE 22
RUN chmod u+s /usr/bin/sudo
RUN adduser -h /home/vagd -s /bin/ash -g sudo -D vagd
RUN echo "vagd ALL=(ALL) NOPASSWD:ALL" > /etc/sudoers.d/vagd && chmod 0440 /etc/sudoers.d/vagd
RUN echo "vagd:vagd" | chpasswd

USER vagd

WORKDIR /home/vagd

COPY keyfile.pub .ssh/authorized_keys

USER root
RUN ssh-keygen -A
RUN mkdir -p /run/sshd && \
    chmod 755 /run/sshd


CMD /usr/sbin/sshd; \
    while true; do sleep 1m; done

'''



#!/bin/env python3
import os

from pwn import *
import vagd.virts.pwngd
from vagd import Vagd, Qegd, Shgd, Dogd, Logd, wrapper, Box

GDB_OFF = 0x555555554000
IP = ''
PORT = 0
BINARY = './bin/sysinfo'
ARGS = []
ENV = {}
API = True
GDB = f"""
b main
c"""

context.binary = exe = ELF(BINARY, checksec=False)
context.aslr = False

byt = lambda x: str(x).encode()

LOCKFILE = vagd.virts.pwngd.Pwngd.LOCKFILE


def test_lockfile(expected):
    with open(LOCKFILE) as lockfile:
        assert lockfile.read() == expected, "bad lockfile"


def virts():
    log.info("Testing Local")
    yield Logd(exe.path)

    log.info("Testing Logging")
    context.log_level = 'error'
    yield Logd(exe.path)
    context.log_level = 'info'

    if args.VAGRANT:

        log.info("Testing Vagrant")

        if os.path.exists(Vagd.VAGRANTFILE_PATH):
            os.system(f"VAGRANT_CWD={Vagd.LOCAL_DIR} vagrant destroy -f")
            os.remove(Vagd.VAGRANTFILE_PATH)

        vm = Vagd(exe.path, vbox=Box.VAGRANT_JAMMY64, packages=['cowsay'], tmp=True, fast=True, ex=True)
        assert vm.is_new, "vm should be new"
        assert vm._ssh.which('cowsay'), "cowsay wasn't installed"
        test_lockfile(Vagd.TYPE)
        yield vm
        vm._ssh.close()

        log.info("Testing Vagrant restore")
        vm = Vagd(exe.path, vbox=Box.VAGRANT_JAMMY64, tmp=True, fast=True, ex=True)
        assert not vm.is_new, "vm shouldn't be new, restored"
        yield vm
        vm._ssh.close()

        log.info("Testing Vagrant restart")
        os.system(f"VAGRANT_CWD={Vagd.LOCAL_DIR} vagrant halt")
        vm = Vagd(exe.path, vbox=Box.VAGRANT_JAMMY64, tmp=True, fast=True, ex=True)
        assert not vm.is_new, "vm shouldn't be new, restarted"
        yield vm
        vm._ssh.close()

        os.system("vagd clean")


    if not args.NODOGD:
      if os.path.exists(Dogd.LOCKFILE):
          os.remove(Dogd.LOCKFILE)
      log.info("Testing Docker for Ubuntu")
      vm = Dogd(exe.path + "_stat", image=Box.DOCKER_NOBLE, packages=['cowsay'], tmp=True, ex=True, fast=True)
      assert vm.is_new, "vm should be new"
      assert vm._ssh.which('cowsay'), "cowsay wasn't installed"
      yield vm
      vm._ssh.close()

      log.info("Testing Docker for Ubuntu restore")
      vm = Dogd(exe.path + "_stat", image=Box.DOCKER_NOBLE, tmp=True, ex=True, fast=True)
      assert not vm.is_new, "vm shouldn't be new, restored"
      yield vm
      vm._ssh.close()

      os.system("vagd clean")
      log.info("Testing Docker for Alpine")
      vm = Dogd(exe.path + "_stat", image=Box.DOCKER_ALPINE_316, tmp=True, ex=True, fast=True)
      assert vm.is_new, "vm should be new"
      yield vm
      vm._ssh.close()

      log.info("Testing Docker for Alpine restore")
      vm = Dogd(exe.path + "_stat", image=Box.DOCKER_ALPINE_316, tmp=True, ex=True, fast=True)
      assert not vm.is_new, "vm shouldn't be new, restored"
      yield vm
      vm._ssh.close()

      os.system("vagd clean")
    log.info("Testing Qemu")
    vm = Qegd(exe.path + "_stat", img=Box.QEMU_NOBLE, tmp=True, packages=['cowsay'], ex=True, fast=True)
    assert vm.is_new, "vm should be new"
    assert vm._ssh.which('cowsay'), "cowsay wasn't installed"
    yield vm
    vm._ssh.close()

    log.info("Testing Qemu restore")
    vm = Qegd(exe.path + "_stat", img=Box.QEMU_NOBLE, tmp=True, ex=True, fast=True)
    assert not vm.is_new, "vm shouldn't be new, restored"
    yield vm
    vm._ssh.close()
    user = vm._user
    port = vm._port

    log.info("Testing SSH")
    yield Shgd(exe.path + "_stat", user=user, port=port, keyfile=vm._ssh.keyfile, tmp=True, ex=True, fast=True)


for virt in virts():
    t = virt.start(argv=ARGS, env=ENV, gdbscript=GDB, api=API)

    sleep(1)
    if args.GDB:
        g = wrapper.GDB(t)
        g.execute('p "PWN"')
        g.execute('c')

    out = b'\n'.join(t.recvlines(3))

    log.info(out.decode())
    t.close()
    os.system('tmux kill-pane')

os.system("vagd clean")
sleep(1)
assert not os.path.exists(LOCKFILE), "lockfile shouldn't exist"

print("Everything executed without errors")



import os
import sys

import pwnlib.log

from vagd.virts.pwngd import Pwngd

_GENERATE_KEYPAIR = 'ssh-keygen -q -t ed25519 -f {keyfile} -N ""'
log = pwnlib.log.getLogger(None)
log.setLevel(pwnlib.log.logging.INFO)


def _ensure_log():
    """
    ensure the correct logger is set
    """
    global log
    if 'pwn' in sys.modules:
        from pwn import log as pwnlog
        log = pwnlog

def info(info: str):
    """
    log info with pwntools
    :param info: info to log
    """
    _ensure_log()
    log.info(info)


def debug(debug: str):
    """
    log debug with pwntools
    :param debug: debug to log
    """
    _ensure_log()
    log.debug(debug)


def warn(warn: str):
    """
    log warn with pwntools
    :param warn: warn to log
    """
    _ensure_log()
    log.warn(warn)


def error(error: str):
    """
    log error with pwntools
    :param error: error to log
    """
    _ensure_log()
    log.error(error)

def progress(progress: str) -> pwnlib.log.Progress:
    """
    log progress with pwntools
    :param progress: progress to log
    """
    _ensure_log()
    return log.progress(progress)


def generate_keypair():
    """
    generate a keypair in .vagd directory
    """
    if not (os.path.exists(Pwngd.KEYFILE) and os.path.exists(Pwngd.KEYFILE + '.pub')):
        info("No Keypair was found. Generating new keypair")
        os.system(_GENERATE_KEYPAIR.format(keyfile=Pwngd.KEYFILE))


def is_port_in_use(port: int) -> bool:
    """
    check if a port is currently used
    :param port: port to check
    :return: if the port is already used
    """
    import socket
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(('localhost', port)) == 0


def first_free_port(start: int = 2222, tries: int = 101) -> int:
    """
    returns first free port starting from start (max increments = tries)
    :param start: start of port search
    :param tries: number of tries to increment ports
    :return: first free port
    """
    for i in range(tries):
        port = start + i
        if not is_port_in_use(port):
            return port

    error(f'No free port inside range {start}-{start + tries}')



import time
from typing import List

import pwnlib.tubes.ssh

from vagd import helper
from vagd.virts.pwngd import Pwngd


class Shgd(Pwngd):
    """ 
    ssh interface for pwntools

    :param binary: binary to execute
    :param user: ssh user
    :param host: ssh hostname
    :param port: ssh port
    :param keyfile: ssh keyfile (default in .vagd)
    :param kwargs: parameters to pass through to super
    """

    DEFAULT_HOST = 'localhost'
    DEFAULT_PORT = 22
    DEFAULT_USER = 'root'

    _user: str
    _host: str
    _port: int
    _keyfile: str
    _ssh: pwnlib.tubes.ssh.ssh

    def bind(self, port: int) -> int:
        """
        bind port from ssh connection locally
        :param port:
        :return:
        """

        remote = self._ssh.connect_remote('127.0.0.1', port)
        listener = pwnlib.tubes.listen.listen(0)
        port = listener.lport

        # Disable showing GDB traffic when debugging verbosity is increased
        remote.level = 'error'
        listener.level = 'error'

        # Hook them up
        remote.connect_both(listener)

        return port

    def _vm_setup(self) -> None:
        """
        pass
        """
        pass

    _TRIES = 3  # three times the charm

    def _ssh_setup(self) -> None:
        """
        setup ssh connection
        """
        progress = helper.progress("connecting to ssh")
        for i in range(Shgd._TRIES):
            try:
                self._ssh = pwnlib.tubes.ssh.ssh(
                    user=self._user,
                    host=self._host,
                    port=self._port,
                    keyfile=self._keyfile,
                    ignore_config=True
                )
                progress.success("Done")
                break
            except:
                if i + 1 == Shgd._TRIES:
                    progress.failure('Failed')
                    helper.error("Failed to connect to ssh")
                else:
                    progress.status('Trying again')
                time.sleep(1 if i == 0 else 10)

    def __init__(self,
                 binary: str,
                 user: str = DEFAULT_USER,
                 host: str = DEFAULT_HOST,
                 port: int = DEFAULT_PORT,
                 keyfile: str = Pwngd.KEYFILE,
                 **kwargs):
        """

        :param binary: binary to execute
        :param user: ssh user
        :param host: ssh hostname
        :param port: ssh port
        :param keyfile: ssh keyfile (default in .vagd)
        :param kwargs: parameters to pass through to super
        """
        self._user = user
        self._host = host
        self._port = port
        self._keyfile = keyfile

        self._ssh_setup()

        super().__init__(binary=binary, **kwargs)



import logging
import os
import pathlib
from abc import ABC, abstractmethod
from shutil import which
from typing import Union, Dict, Iterable, List

import pwnlib.args
import pwnlib.filesystem
import pwnlib.gdb
import pwnlib.tubes

from vagd import helper


class Pwngd(ABC):
    """
    start binary on remote and return pwnlib.tubes.process.process

    :param binary: binary for VM debugging
    :param files: other files or directories that need to be uploaded to VM
    :param packages: packages to install on vm
    :param symbols: additionally install libc6 debug symbols
    :param tmp: if a temporary directory should be created for files
    :param gdbsrvport: specify static gdbserver port, REQURIES port forwarding to localhost
    :param fast: mounts libs locally for faster symbol extraction (experimental)
    :param ex: if experimental features should be enabled
    """
    LOCAL_DIR = './.vagd/'
    HOME_DIR = os.path.expanduser('~/.vagd/')
    SYSROOT = LOCAL_DIR + 'sysroot/'
    LOCKFILE = LOCAL_DIR + 'vagd.lock'
    SYSROOT_LIB = SYSROOT + 'lib/'
    SYSROOT_LIB_DEBUG = SYSROOT + 'lib/debug'
    KEYFILE = HOME_DIR + 'keyfile'
    PUBKEYFILE = KEYFILE + '.pub'
    DEFAULT_PORT = 2222
    STATIC_GDBSRV_PORT = 42069

    is_new: bool = False
    _path: str
    _gdbsrvport: int
    _binary: str
    _ssh: pwnlib.tubes.ssh.ssh
    _experimental: bool
    _fast: bool

    @abstractmethod
    def _vm_setup(self) -> None:
        """
        setup vagrant machine creates new one if no Vagrantfile is specified or box does not match
        """
        pass

    @abstractmethod
    def _ssh_setup(self) -> None:
        """
        setup ssh connection
        """
        pass

    def _sync(self, file: str) -> bool:
        """
        upload file on remote if not exist
        :type file: file to upload
        :return: if the file was uploaded
        """
        sshpath = pwnlib.filesystem.SSHPath(file)
        if not sshpath.exists():
            self.put(file)
            return True
        return False

    _SSHFS_TEMPLATE = \
        'sshfs -p {port} -o StrictHostKeyChecking=no,ro,IdentityFile={keyfile} {user}@{host}:{remote_dir} {local_dir}'

    def _mount(self, remote_dir: str, local_dir: str) -> None:
        """
        mount remote dir on local wiith sshfs
        :param remote_dir: directory on remote to mount
        :param local_dir: local mount point
        """
        if not which('sshfs'):
            self._error('sshfs isn\'t installed')
        cmd = Pwngd._SSHFS_TEMPLATE.format(port=self._ssh.port,
                                           keyfile=self._ssh.keyfile,
                                           user=self._ssh.user,
                                           host=self._ssh.host,
                                           remote_dir=remote_dir,
                                           local_dir=local_dir)
        helper.info(cmd)
        os.system(cmd)

    def _lock(self, typ: str):
        if not os.path.exists(Pwngd.LOCAL_DIR):
            os.makedirs(Pwngd.LOCAL_DIR)

        with open(Pwngd.LOCKFILE, 'w') as lfile:
            lfile.write(typ)

    def _mount_lib(self, remote_lib: str = '/usr/lib') -> None:
        """
        mount the lib directory of remote
        """
        if not (os.path.exists(Pwngd.SYSROOT) and os.path.exists(Pwngd.SYSROOT_LIB)):
            os.makedirs(Pwngd.SYSROOT_LIB)
        if not os.path.ismount(Pwngd.SYSROOT_LIB):
            helper.info('mounting libs in sysroot')
            self._mount(remote_lib, Pwngd.SYSROOT_LIB)

    def system(self, cmd: str) -> pwnlib.tubes.ssh.ssh_channel:
        """
        executes command on vm, interface to  pwnlib.tubes.ssh.ssh.system

        :param cmd: command to execute on vm
        :return: returns
        """
        return self._ssh.system(cmd)

    DEFAULT_PACKAGES = ['gdbserver', 'python3', 'sudo']
    LIBC6_DEBUG = 'libc6-dbg'
    LIBC6_I386 = 'libc6-i386'

    def _install_packages(self, packages: Iterable):
        """
        install packages on remote machine

        :param packages: packages to install on remote machine
        """
        self.system("sudo apt update").recvall()
        packages_str = " ".join(packages)
        self.system(f"sudo NEEDRESTART_MODE=a apt install -y {packages_str}").recvall()

    def put(self, file: str, remote: str = None):
        """
        upload file or dir on vm,

        :param file: file to upload
        :param remote: remote location of file, working directory if not specified
        :return: returns
        """
        if os.path.isdir(file):
            self._ssh.upload_dir(file, remote=remote)
        else:
            self._ssh.upload(file, remote=remote)

    def pull(self, file: str, local: str = None):
        """
        download file or dir on vm,

        :param file: remote location of file, working directory if not specified
        :param local: local location of file, current directory if not specified
        :return: returns
        """
        sshpath = pwnlib.filesystem.SSHPath(os.path.basename(file))
        if sshpath.is_dir():
            self._ssh.download_dir(file, local=local)
        else:
            self._ssh.download_file(file, local=local)

    LIBS_DIRECTORY = "libs"
    def libs(self, directory=None):
        """
        Downloads the libraries referred to by a file.
        This is done by running ldd on the remote server, parsing the output and downloading the relevant files.

        directory(str): Output directory
        :return:
        """
        for lib in self._ssh._libs_remote(self._binary).keys():
            self.pull(lib, directory + '/' + os.path.basename(lib))

    def __init__(self,
                 binary: str,
                 libs=False,
                 files: Union[str, list[str]] = None,
                 packages: List[str] = None,
                 symbols=True,
                 tmp: bool = False,
                 gdbsrvport: int = -1,
                 fast: bool = False,
                 ex: bool = False):
        """
        Default init setups provided ssh machine

        :param binary: binary for VM debugging
        :param libs: download libraries (using ldd) from VM
        :param files: other files or directories that need to be uploaded to VM
        :param packages: packages to install on vm
        :param symbols: additionally install libc6 debug symbols
        :param tmp: if a temporary directory should be created for files
        :param gdbsrvport: specify static gdbserver port, REQURIES port forwarding to localhost
        :param fast: mounts libs locally for faster symbol extraction (experimental)
        :param ex: if experimental features should be enabled
        """

        self._path = binary
        self._gdbsrvport = gdbsrvport
        self._binary = './' + os.path.basename(binary)


        pwnlib.context.context.ssh_session = self._ssh

        if tmp:
            self._ssh.set_working_directory()

        if self._sync(self._path):
            self.system('chmod +x ' + self._binary)

        if self.is_new and libs:
            if not (os.path.exists(Pwngd.LIBS_DIRECTORY)):
                os.makedirs(Pwngd.LIBS_DIRECTORY)

            self.libs(Pwngd.LIBS_DIRECTORY)

        if self.is_new and packages is not None:
            if symbols:
                packages.append(Pwngd.LIBC6_DEBUG)
            try:
                elf = pwnlib.elf.ELF(binary)
                if elf.arch == 'i386':
                    packages.append(Pwngd.LIBC6_I386)
            except:
                helper.warn("failed to get architecture from binary")
            self._install_packages(packages)


        self._fast = fast
        self._experimental = ex

        if self._fast:
            if self._experimental:
                self._mount_lib()
            else:
                helper.error('requires experimental features, activate with ex=True')

        # Copy files to remote
        if isinstance(files, str):
            self._sync(files)
        elif hasattr(files, '__iter__'):
            for file in files:
                self._sync(file)

    def debug(self,
              argv: list[str] = None,
              exe: str = '',
              env: Dict[str, str] = None,
              ssh=None,
              gdbscript: str = '',
              api: bool = False,
              sysroot: str = None,
              gdb_args: list[str] = None,
              **kwargs) -> pwnlib.tubes.process:
        """
        run binary in vm with gdb and experimental features

        :param argv: command line arguments
        :param exe: exe to execute
        :param env: environment variable dictionary
        :param ssh: ignored self._ssh is used instead
        :param gdbscript: used gdbscript
        :param api: return gdb python api interface
        :param sysroot: sysroot directory
        :param gdb_args: additional gdb arguments
        :param kwargs: pwntool arguments
        :rtype: pwnlib.tubes.process.process
        """
        if argv is None:
            argv = []
        args = [self._binary, ] + argv
        if gdb_args is None:
            gdb_args = list()
        helper.warn('using experimental features')
        ssh = self._ssh
        if isinstance(args, (bytes, pwnlib.gdb.six.text_type)):
            args = [args]

        runner = pwnlib.gdb._get_runner(ssh)
        which = pwnlib.gdb._get_which(ssh)

        args, env = pwnlib.gdb.misc.normalize_argv_env(args, env, helper)
        if env:
            env = {bytes(k): bytes(v) for k, v in env}

        args = pwnlib.gdb._gdbserver_args(args=args, which=which, env=env)

        # set static port if wanted
        if self._gdbsrvport != -1:
            for i in range(len(args)):
                if args[i] == 'localhost:0':
                    args[i] = f':{self._gdbsrvport}'
                    break

        # Make sure gdbserver/qemu is installed
        if not which(args[0]):
            helper.error("%s is not installed" % args[0])

        # Start gdbserver/qemu
        # (Note: We override ASLR here for the gdbserver process itself.)
        gdbserver = runner(args, env=env, aslr=1, **kwargs)

        # Set the .executable on the process object.
        gdbserver.executable = exe

        # Find what port we need to connect to
        port = pwnlib.gdb._gdbserver_port(gdbserver, ssh) if self._gdbsrvport == -1 else self._gdbsrvport

        host = '127.0.0.1'

        if self._fast:
            gdb_args += ["-ex", f"set sysroot {pathlib.Path().resolve()}/{Pwngd.SYSROOT}"]
            gdbscript = f"set debug-file-directory {Pwngd.SYSROOT_LIB_DEBUG}\n" + gdbscript
        elif sysroot:
            gdb_args += ["-ex", f"set sysroot {sysroot}"]
            gdbscript = f"set debug-file-directory ./{sysroot}/lib/debug\n" + gdbscript
        else:
            gdbscript = "set debug-file-directory /lib/debug\n" + gdbscript

        gdb_args += ["-ex", f"file -readnow {self._path}"]

        tmp = pwnlib.gdb.attach((host, port), exe=exe, gdbscript=gdbscript,
                             gdb_args=gdb_args, ssh=ssh, api=api, sysroot=sysroot)
        if api:
            _, gdb = tmp
            gdbserver.gdb = gdb

        # gdbserver outputs a message when a client connects
        garbage = gdbserver.recvline(timeout=1)

        # Some versions of gdbserver output an additional message
        garbage2 = gdbserver.recvline_startswith(b"Remote debugging from host ", timeout=2)

        return gdbserver

    def pwn_debug(self, argv: list[str] = None, ssh=None, **kwargs) -> pwnlib.tubes.process.process:
        """
        run binary in vm with gdb (pwnlib feature set)

        :param argv: comandline arguments for binary
        :param ssh: ignored self._ssh is used instead
        :param kwargs: pwntool parameters
        :return: pwntools process
        """
        if argv is None:
            argv = list()
        return pwnlib.gdb.debug([self._binary] + argv, ssh=self._ssh, **kwargs)

    def process(self, argv: list[str] = None, **kwargs) -> pwnlib.tubes.process.process:
        """
        run binary in vm as process

        :param argv: comandline arguments for binary
        :param kwargs: pwntool parameters
        :return: pwntools process
        """
        if argv is None:
            argv = list()
        return self._ssh.process([self._binary] + argv, **kwargs)

    def start(self,
              argv: list[str] = None,
              gdbscript: str = '',
              api: bool = None,
              sysroot: str = None,
              gdb_args: list = None,
              **kwargs) -> pwnlib.tubes.process.process:
        """
        start binary on remote and return pwnlib.tubes.process.process

        :param argv: commandline arguments for binary
        :param gdbscript: GDB script for GDB
        :param api: if GDB API should be enabled (experimental)
        :param sysroot: sysroot dir (experimental)
        :param gdb_args: extra gdb args (experimental)
        :param kwargs: pwntool parameters
        :return: pwntools process, if api=True tuple with gdb api
        """
        if pwnlib.args.args.GDB:
            if self._experimental:
                return self.debug(argv=argv, gdbscript=gdbscript, gdb_args=gdb_args, sysroot=sysroot,
                                  api=api, **kwargs)
            else:
                if gdb_args or sysroot or api:
                    helper.error('requires experimental features, activate with ex=True in constructor')
                return self.pwn_debug(argv=argv, gdbscript=gdbscript, sysroot=sysroot, **kwargs)
        else:
            return self.process(argv=argv, **kwargs)



import os
import time
from shutil import which, copyfile
from typing import Dict, List
from urllib.parse import urlparse

from vagd import helper
from vagd.box import Box
from vagd.virts.pwngd import Pwngd
from vagd.virts.shgd import Shgd
from urllib.request import urlretrieve
import pwnlib


class Qegd(Shgd):
    """
    | QEMU Virtualization for pwntools

    :param binary: binary for VM debugging
    :param img: qemu image to use (requires ssh)
    :param user: user inside qemu image
    :param ports: forwarded ports
    :param packages: packages to install on vm
    :param arm: emulate arm in qemu
    :param qemu: qemu cmd
    :param cpu: value for :code -cpu
    :param memory: value for :code -m
    :param cores: value for :code -smp
    :param machine: value for :code -machine
    :param bios: value for :code -bios
    :param custom: custom qemu arguments
    :param detach: run qemu in new terminal
    :param kwargs: parameters to pass through to super

    | SSH from cmd

    .. code-block:: bash

        vagd ssh
        # or
        ssh -o "StrictHostKeyChecking=no" -i ~/.vagd/keyfile -p $(cat .vagd/qemu.lock) ubuntu@0.0.0.0

    | Kill from cmd:

    .. code-block:: bash

        vagd clean
        # or
        kill $(pgrep qemu)

    | Qemu images are cached in the home directory: :code:`~/.vagd/qemu-imgs/`
    |
    | current used images are stored in the local directory: :code:`./.vagd/current.img`
    | These should be deleted automatically, but if a machine gets improperly stopped
    | (shutdown host while vm is running) it might remain and use up space. You can find remaining images with:

    .. code-block:: bash

        find ~/ -name current.img
        rm <path/current.img>

    """

    DEFAULT_IMG = Box.QEMU_NOBLE
    QEMU_DIR = Pwngd.LOCAL_DIR
    IMGS_DIR = Pwngd.HOME_DIR + 'qemu-imgs/'
    DEFAULT_USER = 'vagd'
    DEFAULT_HOST = '0.0.0.0'
    TYPE = 'qegd'
    DEFAULT_PORT = 2222

    DEFAULT_QEMU_CMD = "qemu-system-x86_64"
    DEFAULT_QEMU_ARM_CMD = "qemu-system-aarch64"

    DEFAULT_QEMU_MACHINE_PREFIX = "-machine"
    DEFAULT_QEMU_MACHINE = "accel=kvm,type=q35"
    DEFAULT_QEMU_ARM_MACHINE = "virt"

    DEFAULT_QEMU_CPU_PREFIX = "-cpu"
    DEFAULT_QEMU_CPU = "host"
    DEFAULT_QEMU_ARM_CPU = "cortex-a72"

    DEFAULT_QEMU_CORES_PREFIX = "-smp"
    DEFAULT_QEMU_CORES = "2"

    DEFAULT_QEMU_BIOS_PREFIX = "-bios"
    DEFAULT_QEMU_ARM_BIOS = "/usr/share/edk2/aarch64/QEMU_EFI.fd"

    DEFAULT_QEMU_MEMORY_PREFIX = "-m"
    DEFAULT_QEMU_MEMORY = "2G"


    _img: str
    _local_img: str
    _user: str
    _host: str
    _port: int
    _forward: Dict[str, int]
    _qemu: str
    _cpu: str
    _cores: str
    _memory: str
    _bios: str
    _cores: str
    _machine: str
    _detach: bool

    @staticmethod
    def _is_local(url) -> bool:
        """
        check if provided url is local or remote
        :param url: url to check
        :return: if the url is local or remote
        """
        url_parsed = urlparse(url)
        if url_parsed.scheme in ('file', ''):  # Possibly a local file
            return os.path.exists(url_parsed.path)
        return False

    CURRENT_IMG = QEMU_DIR + "current.img"

    def _set_local_img(self):
        """
        get local image for qemu machine
        """
        if Qegd._is_local(self._img):
            helper.info("Using local image")
            self._local_img = self._img
        else:
            if not os.path.exists(Qegd.IMGS_DIR):
                os.makedirs(Qegd.IMGS_DIR)
            self._local_img = Qegd.IMGS_DIR + urlparse(self._img).path.rsplit('/', 1)[-1]
            if not os.path.exists(self._local_img):
                helper.info("online qemu image starting download")
                urlretrieve(self._img, self._local_img)
                helper.info(f"saved image to {self._local_img}")
        copyfile(self._local_img, Qegd.CURRENT_IMG)

    METADATA_FILE = QEMU_DIR + 'metadata.yaml'
    _METADATA = """instance-id: iid-local01
local-hostname: cloudimg
"""
    USER_DATA_FILE = QEMU_DIR + 'user-data.yaml'
    _USER_DATA = """#cloud-config
users:
  - default
  - name: vagd
    groups: sudo
    shell: /bin/bash
    sudo: ['ALL=(ALL) NOPASSWD:ALL']
    ssh_authorized_keys:
      - {pubkey}
"""

    SEED_FILE = QEMU_DIR + "seed.img"

    _GENERATE_SEED_IMG = f'cloud-localds {SEED_FILE} {USER_DATA_FILE} {METADATA_FILE}'

    def _setup_seed(self):
        """
        create seed.img with config data like ssh keypair in .qemu
        """
        if not which('cloud-localds'):
            helper.error("cloud-image-utils is not installed")
        if not os.path.exists(Qegd.SEED_FILE):
            helper.info(f"{Qegd.SEED_FILE} not found generating new one")
            if not os.path.exists(Qegd.METADATA_FILE):
                helper.info(f"{Qegd.METADATA_FILE} not found generating new one")
                with open(Qegd.METADATA_FILE, 'w') as metadata_file:
                    metadata_file.write(Qegd._METADATA)
            if not os.path.exists(Qegd.USER_DATA_FILE):
                helper.info(f"{Qegd.USER_DATA_FILE} not found generating new one")
                helper.generate_keypair()
                with open(Qegd.USER_DATA_FILE, 'w') as user_data_file:
                    with open(Pwngd.PUBKEYFILE, 'r') as pubkey_file:
                        pubkey = pubkey_file.readline()
                    user_data_file.write(Qegd._USER_DATA.format(pubkey=pubkey))
            os.system(Qegd._GENERATE_SEED_IMG)

    _QEMU_PORT_FORWARDING = ',hostfwd={type}::{guest}-:{host}'
    _QEMU_START = "{qemu} " \
                  + "{machine} " \
                  + "{cores} " \
                  + "{cpu} " \
                  + "{memory} " \
                  + "-nographic " \
                  + "{bios} " \
                  + "-device virtio-net-pci,netdev=net0 " \
                  + "-netdev user,id=net0,hostfwd=tcp::{port}-:22" \
                  + "{ports} " \
                  + "-drive if=virtio,format=qcow2,file={img} " \
                  + "-drive if=virtio,format=raw,file={seed} " \
                  + "{custom} "

    _QEMU_PIPE = "&> /dev/null; "

    _QEMU_SUFFIX = "rm {lock} {current}"

    _QEMU_ARM_START = ""
    LOCKFILE = QEMU_DIR + "qemu.lock"

    def _qemu_start(self):
        """
        start qemu machine
        """
        self._lock(Qegd.TYPE)
        helper.info(f"starting qemu machine, ssh port {self._port}")
        with open(Qegd.LOCKFILE, 'w') as lockfile:
            lockfile.write(str(self._port))
        port_forwarding = "".join(Qegd._QEMU_PORT_FORWARDING.format(type=host.split('/')[1], host=host.split('/')[0], guest=guest)
                                  for host, guest in self._forward.items())
        qemu_cmd = Qegd._QEMU_START.format(qemu=self._qemu,
                                           machine=f'{Qegd.DEFAULT_QEMU_MACHINE_PREFIX} {self._machine}' if self._machine else '',
                                           cores=f'{Qegd.DEFAULT_QEMU_CORES_PREFIX} {self._cores}' if self._cores else '',
                                           cpu=f'{Qegd.DEFAULT_QEMU_CPU_PREFIX} {self._cpu}' if self._cpu else '',
                                           memory=f'{Qegd.DEFAULT_QEMU_MEMORY_PREFIX} {self._memory}' if self._memory else '',
                                           bios=f'{Qegd.DEFAULT_QEMU_BIOS_PREFIX} {self._bios}' if self._bios else '',
                                           port=self._port,
                                           ports=port_forwarding,
                                           img=Qegd.CURRENT_IMG,
                                           custom=self._custom if self._custom else '',
                                           seed=Qegd.SEED_FILE)
        qemu_suffix =  Qegd._QEMU_SUFFIX.format(lock=f'{Qegd.LOCKFILE} {Pwngd.LOCKFILE}',
                                                current=Qegd.CURRENT_IMG)

        helper.info(qemu_cmd)
        if self._detach:
            pwnlib.util.misc.run_in_new_terminal(qemu_cmd, kill_at_exit=False)
        else:
            pid = os.fork()
            if pid == 0:
                if self._detach:
                    pass
                else:
                    qemu_cmd += Qegd._QEMU_PIPE + qemu_suffix
                    os.system(qemu_cmd)
                exit(0)

    def _new_vm(self) -> None:
        """
        create new vm
        """
        self.is_new = True

        helper.info(f"no Lockfile in {Qegd.LOCKFILE}")
        self._set_local_img()
        self._setup_seed()
        self._port = helper.first_free_port(Qegd.DEFAULT_PORT)
        # start qemu in independent process
        self._qemu_start()
        time.sleep(15)

    def _vm_setup(self) -> None:
        """
        setup qemu machine
        """
        self._host = Qegd.DEFAULT_HOST
        if not os.path.exists(Qegd.LOCKFILE):
            self._new_vm()
        else:
            with open(Qegd.LOCKFILE, 'r') as lockfile:
                self._port = int(lockfile.readline())
            if not helper.is_port_in_use(self._port):
                helper.info(f'Lockfile in {Qegd.LOCKFILE}, but no running machine detected. Creating new one')
                os.remove(Qegd.LOCKFILE)
                self._new_vm()
            helper.info(f'Lockfile in {Qegd.LOCKFILE}. Using running qemu instance at port {self._port}')


    def __init__(self,
                 binary: str,
                 img: str = DEFAULT_IMG,
                 user: str = DEFAULT_USER,
                 forward: Dict[str, int] = None,
                 packages: List[str] = None,
                 arm: bool = False,
                 qemu: str = DEFAULT_QEMU_CMD,
                 cpu: str = DEFAULT_QEMU_CPU,
                 memory: str = DEFAULT_QEMU_MEMORY,
                 machine: str = DEFAULT_QEMU_MACHINE,
                 cores: str = DEFAULT_QEMU_CORES,
                 bios: str = None,
                 detach: bool = False,
                 custom: str = "",
                 **kwargs):
        """

        :param binary: binary for VM debugging
        :param img: qemu image to use (requires ssh)
        :param user: user inside qemu image
        :param forward: Dictionary of forwarded ports, needs to follows format: 'hostport/(tcp|udp)' : guestport
        :param packages: packages to install on vm
        :param arm: emulate arm in qemu
        :param qemu: qemu cmd
        :param cpu: value for :code -cpu
        :param memory: value for :code -m
        :param cores: value for :code -smp
        :param machine: value for :code -machine
        :param bios: value for :code -bios
        :param custom: custom qemu arguments
        :param detach: run qemu in new terminal
        :param kwargs: parameters to pass through to super
        """

        if not which(qemu):
            helper.error(qemu + ' isn\'t installed')

        if packages is None:
            packages = list()

        if not os.path.exists(Qegd.QEMU_DIR):
            helper.info(f"Generating {Qegd.QEMU_DIR} dir")
            os.makedirs(Qegd.QEMU_DIR)

        if arm:
            qemu = Qegd.DEFAULT_QEMU_ARM_CMD if qemu == Qegd.DEFAULT_QEMU_CMD else qemu
            cpu = Qegd.DEFAULT_QEMU_ARM_CPU if cpu == Qegd.DEFAULT_QEMU_CPU else cpu
            machine = Qegd.DEFAULT_QEMU_ARM_MACHINE if machine == Qegd.DEFAULT_QEMU_MACHINE else machine
            bios = Qegd.DEFAULT_QEMU_ARM_BIOS if bios is None else bios

        self._img = img
        self._forward = forward if forward else dict()
        self._qemu = qemu
        self._cpu = cpu
        self._memory = memory
        self._cores = cores
        self._machine = machine
        self._bios = bios
        self._custom = custom
        self._detach = detach

        self._vm_setup()

        packages += Pwngd.DEFAULT_PACKAGES

        super().__init__(binary=binary, user=user, host=self._host, port=self._port, packages=packages, **kwargs)



import importlib.metadata
import os
import stat
import sys
from typing import Optional, Dict, List

import typer
from rich.console import Console
from rich.syntax import Syntax

from vagd import helper
# prevents term.init
from vagd.virts.dogd import Dogd
from vagd.virts.pwngd import Pwngd
from vagd.virts.qegd import Qegd
from vagd.virts.vagd import Vagd

DOGD_BOX = "Box.DOCKER_JAMMY"
DOGD = "vm = Dogd(BINARY, image={box}, {args})  # Docker"
QEGD_BOX = "Box.QEMU_JAMMY"
QEGD = "vm = Qegd(BINARY, img={box}, {args})  # Qemu"
SHGD = "vm = Shgd(BINARY, user='user', host='localhost', port=22, {args})  # SSH"

# deprecated
VAGD_BOX = "Box.VAGRANT_JAMMY64"
VAGD = "vm = Vagd(BINARY, {box}, {args})  # Vagrant"

app = typer.Typer(context_settings={"help_option_names": ["-h", "--help"]})

err_console = Console(stderr=True)
console = Console()


def _version(value: bool) -> None:
    if value:
        version = importlib.metadata.version('vagd')
        typer.echo(f"VAGD v{version}")
        raise typer.Exit()


@app.callback()
def main(
        version: Optional[bool] = typer.Option(
            None,
            "--version",
            "-v",
            help="Show current vagd version and exit.",
            callback=_version,
            is_eager=True,
        ),
) -> None:
    pass


def add_virt(dependencies: List[str],
             vms: List[str],
             dependency: str,
             template: str,
             args: Dict[str, str],
             multi=False,
             box=''):

    dependencies.append(dependency)
    args_str = ', '.join(f'{k}={v}' for k, v in args.items())
    vm = template.format(box=box, args=args_str)
    vms.append(('# ' if multi else '') + vm)


@app.command()
def template(
        binary: Optional[str] = typer.Argument('', help='Binary to Exploit'),
        ip: Optional[str] = typer.Argument('', help='Ip or Domain of the remote target'),
        port: Optional[int] = typer.Argument(0, help='port of the remote target'),
        output_exploit: Optional[bool] = typer.Option(False, '-e',
                                                      help='output file of the template (also add +x) to exploit.py'),
        output: Optional[str] = typer.Option('', '-o',
                                             help='output file of the template (also add +x), default stdout'),
        libc: Optional[str] = typer.Option('', '--libc', '-l', help='add libc to template'),
        libs: Optional[bool] = typer.Option(False, '--libs', help='download libraries from virt'),
        files: Optional[List[str]] = typer.Option([], '--files', '-f', help='add files to remote'),
        aslr: Optional[bool] = typer.Option(False, '--aslr', '-a', help='enable gdb ASLR (default: disabled for gdb)'),
        dogd: Optional[bool] = typer.Option(False, '--dogd', '--docker', '-d', help='create docker template'),
        image: Optional[str] = typer.Option(DOGD_BOX, '--image', help='docker image to use'),
        qegd: Optional[bool] = typer.Option(False, '--qegd', '--qemu', '-q', help='create qemu template'),
        img: Optional[str] = typer.Option(QEGD_BOX, '--img', help='qemu cloud image to use'),
        vagd: Optional[bool] = typer.Option(False, '--vagd', '--vagrant', help='DEPRECATED: create vagrant template'),
        vbox: Optional[str] = typer.Option(VAGD_BOX, '--vbox', help='vagrant box to use'),
        shgd: Optional[bool] = typer.Option(False, '--shgd', '--ssh', '-s', help='create ssh template'),
        local: Optional[bool] = typer.Option(False, '--local', help='create local template'),
):
    """
    creates a template
    """

    if image != DOGD_BOX:
        dogd = True
        image = f"'{image}'"

    if img != QEGD_BOX:
        qegd = True
        img = f"'{img}'"

    if vbox != VAGD_BOX:
        vagd = True
        vbox = f"'{vbox}'"

    templatePath = os.path.dirname(os.path.realpath(__file__))
    templateChunks = []
    aliasesPath = templatePath + "/res/aliases.txt"
    templatePath += '/res/template.txt'
    multi = False
    if not any((dogd, qegd, vagd, shgd)):
        dogd = qegd = True

    if sum((dogd, qegd, vagd, shgd)) > 1:
        multi = True


    env = {}
    if libc:
        files.append(libc)
        env['LD_PRELOAD'] = os.path.basename(libc)

    dependencies = []
    vms = []
    args = dict()

    if libs:
        args['libs'] = True

    if files:
        args['files'] = '[' + ','.join(f"'{file}'" for file in files) + ']'

    args['ex'] = 'True'
    args['fast'] = 'True'

    if dogd:
        add_virt(dependencies, vms, 'Dogd', DOGD, args, box=image)
    if qegd:
        add_virt(dependencies, vms, 'Qegd', QEGD, args, multi, box=img)
    if vagd:
        add_virt(dependencies, vms, 'Vagd', VAGD, args, multi, box=vbox)
    if shgd:
        add_virt(dependencies, vms, 'Shgd', SHGD, args, multi)

    with open(aliasesPath, 'r') as aliases_file:
        aliases = aliases_file.read()


    if libc and libs:
        err_console.print("using --libs and --libc, uncomment libc after init")

    with (open(templatePath, 'r') as templateFile):
        for line in templateFile.readlines():

            if libc and not libs and line.startswith('# libc'):
                templateChunks.append(line[2:])
            else:
                templateChunks.append(line)

        template = ''.join(templateChunks).format('{}',
                                                  aliases=aliases,
                                                  binary=binary,
                                                  ip=ip,
                                                  port=str(port),
                                                  env=env,
                                                  libc=libc,
                                                  aslr=aslr,
                                                  is_local=True if local else 'args.LOCAL',
                                                  dependencies=', '.join(dependencies),
                                                  vms=('\n' + ' ' * 4).join(vms))

        if output_exploit:
            output = 'exploit.py'
        if output:
            with open(output, 'w') as exploitFile:
                exploitFile.write(template)
            current_permissions = os.stat(output).st_mode
            new_permissions = current_permissions | (stat.S_IXUSR)
            os.chmod(output, new_permissions)
        else:
            syntax = Syntax(template, 'python', theme='ansi_dark')
            console.print(syntax)


@app.command()
def info(
        binary: str = typer.Argument(..., help='Binary to analyse'),
):
    """
    analyses the binary, prints checksec and .comment (often includes Distro and Compiler info)
    """
    import pwn
    elf = pwn.ELF(binary)
    helper.info(elf.section('.comment').decode().replace('\0', '\n'))


def _get_type() -> str:
    if os.path.exists(Pwngd.LOCKFILE):
        with open(Pwngd.LOCKFILE) as lfile:
            return lfile.read()
    sys.stderr.write("no vagd instance is running\n")
    exit(1)


def _exec(cmd: str, env: Dict = None):
    if env is None:
        env = os.environ
    else:
        env.update(os.environ)
    os.execvpe("sh", ('sh', '-c', cmd), env)


def _ssh(port, user):
    os.system(
        f'ssh -o "StrictHostKeyChecking=no" -i {Pwngd.KEYFILE} -p {port} {user}@0.0.0.0')


@app.command()
def ssh(
        user: Optional[str] = typer.Option(None, '--user', '-u', help='ssh user'),
):
    """
    ssh to current vagd instance (must be in exploit dir)
    """
    typ = _get_type()
    if typ == Dogd.TYPE:
        if user is None:
            user = Dogd.DEFAULT_USER
        with open(Dogd.LOCKFILE, 'r') as lfile:
            port = lfile.read().split(':')[1]
            _ssh(int(port), user)
    elif typ == Qegd.TYPE:
        if user is None:
            user = Qegd.DEFAULT_USER
        with open(Qegd.LOCKFILE, 'r') as lfile:
            _ssh(int(lfile.read()), user)
    elif typ == Vagd.TYPE:
        os.system(f'VAGRANT_CWD={Pwngd.LOCAL_DIR} vagrant ssh')
    else:
        sys.stderr.write(f"Unknown type in {Pwngd.LOCKFILE}: {typ}\n")
        exit(1)


def _scp(port: int, user: str, source: str, target: str, recursive: bool, keyfile: str = Pwngd.KEYFILE):
    options = '-R' if recursive else ''
    if ':' in source:
        file = source[source.find(':') + 1:]
        source = f'{user}@0.0.0.0:{file}'
    if ':' in target:
        file = target[target.find(':') + 1:]
        target = f'{user}@0.0.0.0:{file}'
    os.system(f'scp -P {port} -o StrictHostKeyChecking=no -i {keyfile} {options} {source} {target}')


@app.command()
def scp(
        source: str = typer.Argument(..., help='source file'),
        target: str = typer.Argument('vagd:./', help='target file'),
        recursive: bool = typer.Option(False, '-r', '--recursive', help='recursive copy'),
        user: Optional[str] = typer.Option(None, '--user', '-u', help='ssh user'),
):
    """
    scp to from current vagd instance (must be in exploit dir)
    """
    typ = _get_type()
    if typ == Dogd.TYPE:
        if user is None:
            user = Dogd.DEFAULT_USER
        with open(Dogd.LOCKFILE, 'r') as lfile:
            port = lfile.read().split(':')[1]
            _scp(int(port), user, source, target, recursive)
    elif typ == Qegd.TYPE:
        if user is None:
            user = Qegd.DEFAULT_USER
        with open(Qegd.LOCKFILE, 'r') as lfile:
            _scp(int(lfile.read()), user, source, target, recursive)
    elif typ == Vagd.TYPE:
        import vagrant
        v = vagrant.Vagrant(os.path.dirname(Vagd.VAGRANTFILE_PATH))
        _scp(2222, v.user(), source, target, recursive, v.keyfile())
    else:
        sys.stderr.write(f"Unknown type in {Pwngd.LOCKFILE}: {typ}\n")
        exit(1)


@app.command()
def clean():
    """
    clean current vagd instance (stop/kill/remove/destroy)
    """
    typ = _get_type()
    if typ == Dogd.TYPE:
        if os.path.exists(Dogd.LOCKFILE):
            with open(Dogd.LOCKFILE, 'r') as lockfile:
                data = lockfile.readline().split(':')
                id = data[0]
            import docker
            client = docker.from_env()
            if not client.containers.list(filters={'id': id}):
                sys.stderr.write(f'Lockfile {Dogd.LOCKFILE} found, container not running\n')
                exit(1)
            else:
                container = client.containers.get(id)
                typer.echo(f'Lockfile {Dogd.LOCKFILE} found, Docker Instance f{container.short_id}')
                container.kill()
    elif typ == Qegd.TYPE:
        os.system("kill $(pgrep qemu)")
    elif typ == Vagd.TYPE:
        import vagrant
        v = vagrant.Vagrant(os.path.dirname(Vagd.VAGRANTFILE_PATH))
        v.halt()
        v.destroy()
        os.remove(Vagd.VAGRANTFILE_PATH)
    else:
        sys.stderr.write(f"Unknown type in {Pwngd.LOCKFILE}: {typ}\n")
        exit(1)
    os.remove(Pwngd.LOCKFILE)


def start():
    app()
